@@ -58,9 +58,10 @@ static int use_gpio_keys = FALSE;
 
 // Accelerometer parameters
 static int accel_enable = FALSE;
-static int accel_deadzone = 150;
-static int accel_center_y = 150;
-static int accel_center_x = 0;
+static int accel_portrait = FALSE;
+static float accel_deadzone = 10;
+static float accel_center_x = 0;
+static float accel_center_y = 0;
 
 // Local key masks
 enum gpsp_key {
@@ -267,7 +268,55 @@ pause_keybd_handler(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
 static void
 activate_tasknav(GtkWidget *widget, gpointer user_data)
 {
-    system("dbus-send /com/nokia/hildon_desktop com.nokia.hildon_desktop.exit_app_view");
+  system("dbus-send /com/nokia/hildon_desktop com.nokia.hildon_desktop.exit_app_view");
+}
+
+static void accel_update_button(GtkWidget *button)
+{
+  char value[32];
+  float pitch = accel_center_y * 180/M_PI;
+  float roll  = accel_center_x * 180/M_PI;
+
+  snprintf(value, sizeof(value), "pitch = %.1f; roll = %.1f;", pitch, roll);
+
+  hildon_button_set_value(HILDON_BUTTON(button), value);
+}
+
+static void accel_read(float *ret_x, float *ret_y)
+{
+  // Read the accelerometer
+  float x, y, z;
+  FILE* accel = fopen("/sys/class/i2c-adapter/i2c-3/3-001d/coord", "r");
+  if (accel) {
+    if (accel_portrait) {
+      fscanf(accel, "%f %f %f", &y, &x, &z);
+      x = -x;
+    } else {
+      fscanf(accel, "%f %f %f", &x, &y, &z);
+    }
+    fclose(accel);
+  } else {
+    fprintf (stderr, "CRITICAL: cannot access the accelerometer\n");
+    exit(1);
+  }
+
+  // Convert to angular coordinates
+  *ret_x = atan2(x, sqrt(y*y + z*z)) + accel_center_x;
+  *ret_y = fmod(atan2(y, -z) + accel_center_y + M_PI, 2*M_PI) - M_PI;
+}
+
+static void
+accel_calibrate(GtkWidget *widget, gpointer user_data)
+{
+  accel_center_x = 0;
+  accel_center_y = 0;
+
+  accel_read(&accel_center_x, &accel_center_y);
+  accel_center_x = -accel_center_x;
+  accel_center_y = -accel_center_y;
+
+  if (widget)
+    accel_update_button(widget);
 }
 
 static void activate_pause(int automatic)
@@ -296,13 +345,38 @@ static void activate_pause(int automatic)
     gtk_button_set_label (GTK_BUTTON (tasknav_button), "Tasks");
     g_signal_connect (tasknav_button, "clicked", G_CALLBACK (activate_tasknav), NULL);
 
+    // Create accelerometer button
+    GtkWidget *accel_button = hildon_check_button_new (HILDON_SIZE_FINGER_HEIGHT);
+    gtk_button_set_label (GTK_BUTTON (accel_button), "Enable");
+    hildon_check_button_set_active (HILDON_CHECK_BUTTON (accel_button), accel_enable);
+
+    // Create accelerometer calibration button
+    GtkWidget *calibrate_button = hildon_button_new (HILDON_SIZE_FINGER_HEIGHT, HILDON_BUTTON_ARRANGEMENT_VERTICAL);
+    hildon_button_set_title (HILDON_BUTTON (calibrate_button), "Calibrate");
+    hildon_button_set_title_alignment (HILDON_BUTTON (calibrate_button), 0.5, 0);
+    accel_update_button (calibrate_button);
+    g_signal_connect (calibrate_button, "clicked", G_CALLBACK (accel_calibrate), NULL);
+
     // Populate the dialog
-    GtkWidget *hbox = gtk_hbox_new (TRUE, 0);
+    GtkWidget *vbox = gtk_vbox_new (FALSE, 8);
+    GtkWidget *other_box = gtk_hbox_new (TRUE, 0);
+    GtkWidget *accel_box = gtk_hbox_new (TRUE, 0);
+    GtkWidget *other_frame = gtk_frame_new ("General");
+    GtkWidget *accel_frame = gtk_frame_new ("Accelerometer");
     GtkWidget *content_area = gtk_dialog_get_content_area (GTK_DIALOG (pause_dialog));
-    gtk_container_add (GTK_CONTAINER (hbox), mute_button);
-    gtk_container_add (GTK_CONTAINER (hbox), tasknav_button);
-    gtk_container_add (GTK_CONTAINER (content_area), hbox);
-    gtk_widget_show_all(hbox);
+
+    gtk_container_add (GTK_CONTAINER (other_box), mute_button);
+    gtk_container_add (GTK_CONTAINER (other_box), tasknav_button);
+    gtk_container_add (GTK_CONTAINER (accel_box), accel_button);
+    gtk_container_add (GTK_CONTAINER (accel_box), calibrate_button);
+
+    gtk_container_add (GTK_CONTAINER (other_frame), other_box);
+    gtk_container_add (GTK_CONTAINER (accel_frame), accel_box);
+
+    gtk_container_add (GTK_CONTAINER (vbox), other_frame);
+    gtk_container_add (GTK_CONTAINER (vbox), accel_frame);
+    gtk_container_add (GTK_CONTAINER (content_area), vbox);
+    gtk_widget_show_all (vbox);
 
     g_signal_connect (G_OBJECT (pause_dialog), "key-release-event", G_CALLBACK (pause_keybd_handler), NULL);
 
@@ -310,6 +384,8 @@ static void activate_pause(int automatic)
       quit();
 
     global_enable_audio = !hildon_check_button_get_active (HILDON_CHECK_BUTTON (mute_button));
+
+    accel_enable = hildon_check_button_get_active (HILDON_CHECK_BUTTON (accel_button));
   }
 
   gtk_widget_destroy (GTK_WIDGET(pause_dialog));
@@ -394,11 +470,13 @@ configure_handler(GtkWidget *widget, GdkEventConfigure *event, gpointer user_dat
 
   if (event->width > event->height) {
     // Landscape
+    accel_portrait = FALSE;
     screen_h = event->height;
     regions = regions_hor;
     hildon_animation_actor_set_show (keypad_actor, FALSE);
   } else {
     // Portrait
+    accel_portrait = TRUE;
     screen_h = event->height - 480; // make some room for the virtual keyapad
     regions = regions_ver;
     hildon_animation_actor_set_show (keypad_actor, TRUE);
@@ -415,22 +493,26 @@ delete_handler()
 
 static void print_help()
 {
-  printf("gpSP - Game Boy Advance emulator"                                         "\n"
-         "Usage: gpspm FILE [options]"                                              "\n"
-         "       gpspm [options]"                                                   "\n"
-         "  --rom FILE          - the ROM file to open"                             "\n"
-         "  --keys FILE         - read key bindings from nonstandard location"      "\n"
-         "  --accelerometer     - enable accelerometer input"                       "\n"
-         "  --autopause         - pause emulation when losing focus"                "\n"
-         "  --scale N           - none (0), double (1), triple (2), fullscreen (3)" "\n"
-         "  --filter N          - none (0), scale2x (1), scale3x (2), eagle2x (3)"  "\n"
-         "  --frameskip-type N  - disabled (0), auto (1), manual (2)"               "\n"
-         "  --frameskip-value N - how many frames should be skipped"                "\n"
-         "  --random-frameskip  - some randomness in choosing which frames to skip" "\n"
-         "  --no-sound          - disable sound"                                    "\n"
-         "  --sound-buffer N    - size of the audio buffer, 2^(4+N)"                "\n"
-         "  --breakpoint N      - set the breakpoint to N"                          "\n"
-         "  --help              - display this message"                             "\n");
+  printf("gpSP - Game Boy Advance emulator"                                                "\n"
+         "Usage: gpspm FILE [options]"                                                     "\n"
+         "       gpspm [options]"                                                          "\n"
+         "  --rom FILE                 - the ROM file to open"                             "\n"
+         "  --keys FILE                - read key bindings from nonstandard location"      "\n"
+         "  --accelerometer            - enable accelerometer input"                       "\n"
+         "  --accelerometer-deadzone A - accelerometer deadzone (degrees)"                 "\n"
+         "  --accelerometer-pitch A    - accelerometer neutral pitch (degrees)"            "\n"
+         "  --accelerometer-roll A     - accelerometer neutral roll (degrees)"             "\n"
+         "  --accelerometer-calibrate  - set neutral pitch and roll on startup"            "\n"
+         "  --autopause                - pause emulation when losing focus"                "\n"
+         "  --scale N                  - none (0), double (1), triple (2), fullscreen (3)" "\n"
+         "  --filter N                 - none (0), scale2x (1), scale3x (2), eagle2x (3)"  "\n"
+         "  --frameskip-type N         - disabled (0), auto (1), manual (2)"               "\n"
+         "  --frameskip-value N        - how many frames should be skipped"                "\n"
+         "  --random-frameskip         - some randomness in choosing which frames to skip" "\n"
+         "  --no-sound                 - disable sound"                                    "\n"
+         "  --sound-buffer N           - size of the audio buffer, 2^(4+N)"                "\n"
+         "  --breakpoint N             - set the breakpoint to N"                          "\n"
+         "  --help                     - display this message"                             "\n");
   exit(1);
 }
 
@@ -492,6 +574,7 @@ static void read_keys(FILE *keys_file)
 
 #define setaopt(optname) if (++i < *argc) optname = argv[i]; else print_help()
 #define setiopt(optname) if (++i < *argc) optname = atoi(argv[i]); else print_help()
+#define setfopt(optname) if (++i < *argc) optname = atof(argv[i]); else print_help()
 
 int gpsp_plat_args(int *argc, char *argv[])
 {
@@ -512,7 +595,11 @@ int gpsp_plat_args(int *argc, char *argv[])
   for (i = 1; i < *argc; i++) {
          if (!strcmp(argv[i], "--rom"))              setaopt(rom_path);
     else if (!strcmp(argv[i], "--keys"))             setaopt(keys_path);
-    else if (!strcmp(argv[i], "--accelerometer"))    accel_enable = TRUE;
+    else if (!strcmp(argv[i], "--accelerometer"))           accel_enable = TRUE;
+    else if (!strcmp(argv[i], "--accelerometer-deadzone"))  { setfopt(accel_deadzone); accel_deadzone /= 180/M_PI; }
+    else if (!strcmp(argv[i], "--accelerometer-pitch"))     { setfopt(accel_center_y); accel_center_y /= 180/M_PI; }
+    else if (!strcmp(argv[i], "--accelerometer-roll"))      { setfopt(accel_center_x); accel_center_x /= 180/M_PI; }
+    else if (!strcmp(argv[i], "--accelerometer-calibrate")) accel_calibrate(NULL, NULL);
     else if (!strcmp(argv[i], "--autopause"))        autopause = TRUE;
     else if (!strcmp(argv[i], "--scale"))            setiopt(screen_scale);
     else if (!strcmp(argv[i], "--filter"))           setiopt(screen_filter2);
@@ -729,15 +816,35 @@ inline u32 gpsp_plat_joystick_read(void)
 {
   int keystate = keybd_keystate | accel_keystate | touch_keystate;
 
-  // Expand keys that translate to multiple basic keys
-  if (combo_keystate & GKEY_LEFT_UP)
-    keystate |= GKEY_LEFT | GKEY_UP;
-  if (combo_keystate & GKEY_RIGHT_UP)
-    keystate |= GKEY_RIGHT | GKEY_UP;
-  if (combo_keystate & GKEY_LEFT_DOWN)
-    keystate |= GKEY_LEFT | GKEY_DOWN;
-  if (combo_keystate & GKEY_RIGHT_DOWN)
-    keystate |= GKEY_RIGHT | GKEY_DOWN;
+  // Allow diagonals to cancel each other out
+  int combo_hor = 0;
+  int combo_ver = 0;
+
+  if (combo_keystate & GKEY_LEFT_UP) {
+    combo_hor--;
+    combo_ver--;
+  }
+  if (combo_keystate & GKEY_RIGHT_UP) {
+    combo_hor++;
+    combo_ver--;
+  }
+  if (combo_keystate & GKEY_LEFT_DOWN) {
+    combo_hor--;
+    combo_ver++;
+  }
+  if (combo_keystate & GKEY_RIGHT_DOWN) {
+    combo_hor++;
+    combo_ver++;
+  }
+
+  if (combo_hor < 0)
+    keystate |= GKEY_LEFT;
+  if (combo_hor > 0)
+    keystate |= GKEY_RIGHT;
+  if (combo_ver < 0)
+    keystate |= GKEY_UP;
+  if (combo_ver > 0)
+    keystate |= GKEY_DOWN;
 
   return keystate;
 }
@@ -756,32 +863,23 @@ void *fb_flip_screen(void)
   if (accel_enable) {
     accel_keystate = 0;
 
-    // Read the accelerometer
-    int x, y, z;
-    FILE* accel = fopen("/sys/class/i2c-adapter/i2c-3/3-001d/coord", "r");
-    if (accel) {
-      fscanf(accel, "%d %d %d", &x, &y, &z);
-      fclose(accel);
-    } else {
-      fprintf (stderr, "CRITICAL: cannot access the accelerometer\n");
-      exit(1);
+    float x, y;
+    accel_read(&x, &y);
+
+    // Convert to key state
+    if (x*x + y*y > accel_deadzone*accel_deadzone) {
+      float direction = atan2(y, x);
+
+      if (direction < M_PI*3/8 && direction > -M_PI*3/8)
+        accel_keystate |= GKEY_LEFT;
+      else if (direction > M_PI*5/8 || direction < -M_PI*5/8)
+        accel_keystate |= GKEY_RIGHT;
+
+      if (direction > M_PI*1/8 && direction < M_PI*7/8)
+        accel_keystate |= GKEY_UP;
+      else if (direction < -M_PI*1/8 && direction > -M_PI*7/8)
+        accel_keystate |= GKEY_DOWN;
     }
-
-    // Apply calibration
-    x += accel_center_x;
-    y += accel_center_y;
-
-    // X axis
-    if (x > accel_deadzone)
-      accel_keystate = GKEY_LEFT;
-    else if (x < -accel_deadzone)
-      accel_keystate = GKEY_RIGHT;
-
-    // Y axis
-    if (y > accel_deadzone)
-      accel_keystate |= GKEY_UP;
-    else if (y < -accel_deadzone)
-      accel_keystate |= GKEY_DOWN;
   }
 
   switch (sw_filter) {
